/*
 *  Clock Driver
 *
 *  This driver initailizes timer1 on the MCF5206 as the 
 *  main system clock 
 *
 *  Author:
 *    David Fiddes, D.J@fiddes.surfaid.org
 *    http://www.calm.hw.ac.uk/davidf/coldfire/
 *
 *  COPYRIGHT (c) 1989-1998.
 *  On-Line Applications Research Corporation (OAR).
 *  Copyright assigned to U.S. Government, 1994.
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 *
 *  http://www.OARcorp.com/rtems/license.html.
 *
 *  $Id: ckinit.c,v 1.6 1998/02/17 23:44:45 joel Exp $
 */

#include <stdlib.h>
#include <bsp.h>
#include <rtems/libio.h>


/*
 * Clock_driver_ticks is a monotonically increasing counter of the
 * number of clock ticks since the driver was initialized.
 */
volatile rtems_unsigned32 Clock_driver_ticks;


/*
 * These are set by clock driver during its init
 */
 
rtems_device_major_number rtems_clock_major = ~0;
rtems_device_minor_number rtems_clock_minor;


/***************************************************************************
   Function : Clock_isr

   Description : This handles the timer interrupt by clearing the timer's
   interrupt flag and announcing the clock tick to the system.
 ***************************************************************************/
rtems_isr
Clock_isr (rtems_vector_number vector)
{
  /* Clear pending interrupt... */
  sim->timer1.ter = MCF5206_TIMER_TER_REF | MCF5206_TIMER_TER_CAP;

  /* Announce the clock tick */
  Clock_driver_ticks++;
  rtems_clock_tick();
}


/***************************************************************************
   Function : Clock_exit

   Description : This shuts down the timer if it was enabled and removes it
   from the MCF5206 interrupt mask.
 ***************************************************************************/
void
Clock_exit (void)
{
  if (BSP_Configuration.ticks_per_timeslice ) {

    /* disable all timer1 interrupts */
    sim->imr = sim->imr | MCF5206_SIM_IMR_T1;   

    /* reset timer1... */
    sim->timer1.tmr = MCF5206_TIMER_TMR_CLK_STOP; 

    /* clear pending */
    sim->timer1.ter = MCF5206_TIMER_TER_REF | MCF5206_TIMER_TER_CAP; 
  }
}


/***************************************************************************
   Function : Install_clock

   Description : This initialises timer1 with the BSP timeslice config value
   as a reference and sets up the interrupt handler for clock ticks.
 ***************************************************************************/
static void
Install_clock (rtems_isr_entry clock_isr)
{
  Clock_driver_ticks = 0;
  if ( BSP_Configuration.ticks_per_timeslice ) {
   
    /* Configure timer1 interrupts */
    sim->icr9 = MCF5206_SIM_ICR_AVEC                   |
                MCF5206_SIM_ICR_IL(TIMER1_IRQ_LEVEL)   |
                MCF5206_SIM_ICR_IP(TIMER1_IRQ_PRIORITY); 

    /* Register the interrupt handler */
    set_vector( Clock_isr, TIMER1_IRQ_VECTOR, 1);
    
    /* Enable timer 1 interrupts */
    sim->imr = sim->imr & (~MCF5206_SIM_IMR_T1);  

    /* Reset timer1 */
    sim->timer1.tmr = MCF5206_TIMER_TMR_RST;    
    sim->timer1.tmr = MCF5206_TIMER_TMR_CLK_STOP;   
    sim->timer1.tmr = MCF5206_TIMER_TMR_RST;    
    sim->timer1.tcn = 0; /* zero count */
    sim->timer1.ter = MCF5206_TIMER_TER_REF | MCF5206_TIMER_TER_CAP; /* clear pending */
    
          
    /*
     * Prescale Timer1 by 1/sysclk so that it counts in microsecs
     * for the SBC5206 as supplied this will be 1/25
     */
    sim->timer1.tmr = MCF5206_TIMER_TMR_PS( MCF5206_CLOCK_SPEED/1000000 )  |
                      MCF5206_TIMER_TMR_CE_NONE |
                      MCF5206_TIMER_TMR_ORI     |
                      MCF5206_TIMER_TMR_FRR     |
                      MCF5206_TIMER_TMR_RST;

    /* Set the timer timeout value from the BSP config */      
    sim->timer1.trr = BSP_Configuration.microseconds_per_tick - 1;
    
    /* Kick it into life! */
    sim->timer1.tmr = sim->timer1.tmr | MCF5206_TIMER_TMR_CLK_MSCLK;
    
    /* Register the driver exit procedure so we can shutdown */
    atexit (Clock_exit);
  }
}


/***************************************************************************
   Function : Clock_initialise

   Description : This is called to setup the clock driver. It calls the 
   hardware setup function and make the device major/minor values available
   for others.
 ***************************************************************************/
rtems_device_driver
Clock_initialize(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void *pargp
)
{
  Install_clock (Clock_isr);
 
  /* Make major/minor avail to others such as shared memory driver */
  rtems_clock_major = major;
  rtems_clock_minor = minor;
 
  return RTEMS_SUCCESSFUL;
}
 

/***************************************************************************
   Function : Clock_control

   Description : This is called to pass control information to the clock 
   driver. At the moment this just runs the interrupt handler or 
   re-registers the interrupt handler on request.
 ***************************************************************************/
rtems_device_driver Clock_control(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void *pargp
)
{
  rtems_unsigned32 isrlevel;
  rtems_libio_ioctl_args_t *args = pargp;

  if (args) {
    /*
     * This is hokey, but until we get a defined interface
     * to do this, it will just be this simple...
     */
    if (args->command == rtems_build_name('I', 'S', 'R', ' ')) {
      Clock_isr( TIMER1_IRQ_VECTOR );
    }
    else if (args->command == rtems_build_name('N', 'E', 'W', ' ')) {
      rtems_interrupt_disable( isrlevel );
       (void) set_vector( args->buffer, TIMER1_IRQ_VECTOR, 1 );
      rtems_interrupt_enable( isrlevel );
    }
  }
  return RTEMS_SUCCESSFUL;
}
