/*
 *  RTEMS/FreeBSD driver for NS8390 Ethernet Chipset
 *
 *  This is a generic NS8390 network driver and requires a mini-driver
 *  to provide board/system specific functionality.
 *
 *  Author:
 *    David Fiddes, D.J@fiddes.surfaid.org
 *    http://www.calm.hw.ac.uk/davidf/coldfire/
 *
 *  COPYRIGHT (c) 1989-1998.
 *  On-Line Applications Research Corporation (OAR).
 *  Copyright assigned to U.S. Government, 1994.
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 * 
 *  http://www.OARcorp.com/rtems/license.html.
 * 
 *  $Id: ns8390.c,v 1.11 1998/02/19 23:02:02 joel Exp $
 */

#include <bsp.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdarg.h>
#include <rtems/error.h>
#include <rtems/rtems_bsdnet.h>

#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/sockio.h>

#include <net/if.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>

#include <ns8390.h>


static const char *version = 
  "NS8390 RTEMS/FreeBSD sub-driver A0.01 22/08/98 David Fiddes (D.J@fiddes.surfaid.org)\n";

/* 0 = no debug, 1 = standard traces, 2 = register accesses, 3 = all accesses inc data */

#define DEBUG_NONE    0
#define DEBUG_STD     1
#define DEBUG_REG     2
#define DEBUG_DATA    3

static const int debug_level = DEBUG_NONE;


/*
 * Number of NS8390s supported by this driver
 */
#define NUM_NS8390_DEVICES  1 /* we'll keep things very simple for now */

/*
 * RTEMS event used by interrupt handler to signal driver tasks.
 * This must not be any of the events used by the network task synchronization.
 */
#define INTERRUPT_EVENT RTEMS_EVENT_1

/*
 * RTEMS event used to start transmit daemon.
 * This must not be the same as INTERRUPT_EVENT.
 */
#define START_TRANSMIT_EVENT  RTEMS_EVENT_2

/*
 * Receive buffer size -- Allow for a full ethernet packet including CRC
 */
#define RBUF_SIZE 1520

#if (MCLBYTES < RBUF_SIZE)
# error "Driver must have MCLBYTES > RBUF_SIZE"
#endif

/*
 * Basic Ethernet constants
 */
#define ET_MINLEN 60

/* Interrupts we want to handle from the device.  */

#define NS8390_INTERRUPTS \
  (NS8390_ISR_PRX | NS8390_ISR_PTX | NS8390_ISR_RXE | NS8390_ISR_TXE | NS8390_ISR_OVW | NS8390_ISR_CNT)


/*
 * Hardware-specific storage
 */

static struct ns8390_device ns8390_devices[NUM_NS8390_DEVICES]; 


/*
 * Module prototypes
 */
static void
ns8390_stop (struct ns8390_device *dp);

void
ns8390_tx_daemon (void *arg);

/* static void
ns8390_sendpacket (struct ifnet *ifp, struct mbuf *m); */

static void
ns8390_rx_daemon (void *arg);

/*static rtems_isr
ns8390_interrupt_handler (rtems_vector_number v); */

static void
ns8390_start (struct ifnet *ifp);


static void
ns8390_init (void *arg);

static void
ns8390_stats (struct ns8390_device *dp);

void
ns8390_init_hardware( struct ns8390_device *dp );

void
ns8390_startup( struct ns8390_device *dp );



/* 
 * Find the NS8390 device which is attached at a particular interrupt
 * vector.
 */
static struct ns8390_device *
ns8390_device_for_irno (int irno)
{
  int i;

  for (i = 0; i < NUM_NS8390_DEVICES; ++i)
  {
    if ( ns8390_devices[i].irno == irno &&
         ns8390_devices[i].arpcom.ac_if.if_softc != NULL )
      return &ns8390_devices[i];
  }

  return NULL;
}


/* 
 * Handle the current NS8390 status.  This is called when the device
 * signals an interrupt.  It is also called at other times while
 * NS8390 interrupts have been disabled.
 */

static void
ns8390_check_status (struct ns8390_device *dp)
{
  unsigned32    ns8390_port;

  setRegister_f setReg;
  getRegister_f getReg;
  
  unsigned char status;

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg      = dp->setRegister;
  getReg      = dp->getRegister;


  /* It seems that we need to use a loop here, because if the NE2000
     signals an interrupt because packet transmission is complete, and
     then receives a packet while interrupts are disabled, it seems to
     sometimes fail to signal the interrupt for the received packet
     when interrupts are reenabled.  (Based on the behaviour of the
     Realtek 8019AS chip).  */
  /* TODO: Determine the heritage of this statement - DJF */

  while (1) {
    status = (*getReg)( ns8390_port, NS8390_PAGE0_READ_ISR );
      
    if ((status & ~(NS8390_ISR_RDC | NS8390_ISR_RST)) == 0)
      break;

#ifdef DEBUG_NE2000
    printk ("NE2000 status 0x%x (8259 enabled: %s; mask: %x)\n", status,
	    i8259s_cache & (1 << sc->irno) ? "no" : "yes",
	    i8259s_cache);
#endif

    /* Check for incoming packet overwrite.  */
    if (status & NS8390_ISR_OVW) {
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_OVW );
      dp->overrun = 1;
      rtems_event_send (dp->rx_daemon_tid, INTERRUPT_EVENT);
    }

    /* Check for transmitted packet.  The transmit daemon may now be
       able to send another packet to the device.  */
    if ((status & (NS8390_ISR_PTX | NS8390_ISR_TXE)) != 0) {
      ++dp->eth_stats.tx_acks;
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, status & (NS8390_ISR_PTX | NS8390_ISR_TXE) );
      --dp->inuse;
      dp->transmitting = 0;
      if (dp->inuse > 0 || (dp->arpcom.ac_if.if_flags & IFF_OACTIVE) != 0)
	rtems_event_send (dp->tx_daemon_tid, START_TRANSMIT_EVENT);
    }

    /* Check for received packet.  */
    if ((status & (NS8390_ISR_PRX | NS8390_ISR_RXE)) != 0) {
      ++dp->eth_stats.rx_acks;
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, status & (NS8390_ISR_PRX | NS8390_ISR_RXE) );
      rtems_event_send (dp->rx_daemon_tid, INTERRUPT_EVENT);
    }

    /* Check for counter change.  */
    if ((status & NS8390_ISR_CNT) != 0) {

      dp->eth_stats.rx_frame_errors  += (*getReg)( ns8390_port, NS8390_PAGE0_READ_CNTR0 );
      dp->eth_stats.rx_crc_errors    += (*getReg)( ns8390_port, NS8390_PAGE0_READ_CNTR1 );
      dp->eth_stats.rx_missed_errors += (*getReg)( ns8390_port, NS8390_PAGE0_READ_CNTR2 );
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_CNT );
    }
  }

  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STA );
}


/* Handle an NS8390 interrupt.  */

rtems_isr
ns8390_interrupt_handler (rtems_vector_number v)
{
  struct ns8390_device *dp;

#if 0
  dp = ns8390_device_for_irno (v);
  if (dp == NULL)
    return;
#else
  dp = &ns8390_devices[0];
#endif

  ++dp->eth_stats.interrupts;

  ns8390_check_status (dp);
}


/*
 * Device functions
 */

/* Stop an NS8390 */

static void
ns8390_stop (struct ns8390_device *dp)
{
  unsigned32    ns8390_port;

  setRegister_f setReg;
  getRegister_f getReg;

  int           i;

  if (debug_level >= DEBUG_STD)
    printf("ns8390_stop() - start\n");

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg      = dp->setRegister;
  getReg      = dp->getRegister;

  dp->arpcom.ac_if.if_flags &= ~IFF_RUNNING;

  /* Stop everything.  */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_RD_AB | NS8390_CR_STP );

  /* Wait for the interface to stop, using I as a time limit.  */
  for (i = 0; i < 5000; ++i)
    {
      unsigned char status;

      status = (*getReg)( ns8390_port, NS8390_PAGE0_READ_ISR );
      if ((status & NS8390_ISR_RST) != 0)
	break;
    }

  dp->inuse = 0;
  dp->nextavail = 0;
  dp->nextsend = 0;
  dp->transmitting = 0;

  if (debug_level >= DEBUG_STD)
    printf("ns8390_stop() - end\n");

}


/*
 * Driver transmit daemon
 */

/* Load an NE2000 packet onto the device.  */

static void
ns8390_loadpacket (struct ns8390_device *dp, struct mbuf *m)
{
  unsigned32    ns8390_port;

  setRegister_f setReg;
  getRegister_f getReg;
  setData_f     setData;

  struct mbuf   *mhold = m;
  int           leftover;
  unsigned char leftover_data;
  int           timeout;

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg      = dp->setRegister;
  getReg      = dp->getRegister;
  setData     = dp->setData;

#ifdef DEBUG_NE2000
  printk ("Uploading NE2000 packet\n");
#endif

  /* Reset remote DMA complete flag.  */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_RDC );

  /* Write out the count.  */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RBCR0, m->m_pkthdr.len );
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RBCR1, m->m_pkthdr.len >> 8 );

  dp->sendlen[dp->nextavail] = m->m_pkthdr.len;

  /* Tell the device which address we want to write to.  */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RSAR0, 0 );
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RSAR1, 
                          NS8390_FIRST_TX_PAGE + dp->nextavail * NS8390_TX_PAGES );
  
  /* Set up the write.  */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_WR | NS8390_CR_STA );

  /* Transfer the mbuf chain to device memory.  NE2000 devices require
     that the data be transferred as words, so we need to handle odd
     length mbufs.  */

  leftover = 0;
  leftover_data = '\0';

  for (; m != NULL; m = m->m_next) {
    int len;
    unsigned char *data;

    len = m->m_len;
    if (len == 0)
      continue;

    data = mtod (m, unsigned char *);

    if (leftover) {
      unsigned char next;

      /* Data left over from previous mbuf in chain.  */
      next = *data++;
      --len;
      (*setData)( ns8390_port, (leftover_data << 8) | next );
      leftover = 0;
    }

    while (len > 1) {
      (*setData)( ns8390_port, (data[0] << 8) | data[1] );
      data += 2;
      len -= 2;
    }

    if (len > 0)
    {
      leftover = 1;
      leftover_data = *data++;
    }
  }

  if (leftover)
    (*setData)( ns8390_port, leftover_data << 8);

  m_freem (mhold);

  /* Wait for the device to complete accepting the data, with a
     limiting counter so that we don't wait too long.  */
  for (timeout = 0; timeout < 100; ++timeout)
  {
    unsigned char status;

    status = (*getReg)( ns8390_port, NS8390_PAGE0_READ_ISR );

#ifdef DEBUG_NE2000
      if ((status &~ MSK_RDC) != 0)
	printk ("Status 0x%x while waiting for acknowledgement of uploaded packet\n",
		status);
#endif

    if ((status & NS8390_ISR_RDC) != 0) 
    {
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_RDC );
      break;
    }
  }

  if (timeout >= 100)
    printf ("Timed out waiting for acknowledgement of uploaded NE2000 packet\n");

  ++dp->nextavail;
  if (dp->nextavail == NS8390_TX_BUFS)
    dp->nextavail = 0;
}

/* Tell the NS8390 to transmit a buffer whose contents we have already
   loaded onto the device.  */

static void
ns8390_transmit (struct ns8390_device *dp)
{
  int len;

  unsigned32            ns8390_port;

  setRegister_f         setReg;
  getRegister_f         getReg;
  getData_f             getData;

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg      = dp->setRegister;
  getReg      = dp->getRegister;
  getData     = dp->getData;

#ifdef DEBUG_NE2000
  printk ("Transmitting NE2000 packet\n");
#endif

  len = dp->sendlen[dp->nextsend];
  if (len < ET_MINLEN)
    len = ET_MINLEN;

  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TBCR0, len & 0xFF );
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TBCR1, (len >> 8) & 0xFF );

  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TPSR, NS8390_FIRST_TX_PAGE + (dp->nextsend * NS8390_TX_PAGES) );

  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_TXP | NS8390_CR_STA );

  ++dp->nextsend;
  if (dp->nextsend == NS8390_TX_BUFS)
    dp->nextsend = 0;

  ++dp->eth_stats.tx_packets;
}


/* The NS8390 packet transmit daemon.  This task is started when the
   NS8390 driver is initialized.  */

void
ns8390_tx_daemon (void *arg)
{
  struct ns8390_device  *dp = (struct ns8390_device *)arg;
  struct ifnet          *ifp = &dp->arpcom.ac_if;

  unsigned32            ns8390_port;

  setRegister_f         setReg;
  getRegister_f         getReg;

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg      = dp->setRegister;
  getReg      = dp->getRegister;

  if (debug_level >= DEBUG_STD)
    printf( "ns8390_tx_daemon() - start\n" );

  while (1) {
    rtems_event_set events;

    /* Wait for a packet to be ready for sending, or for there to be
       room for another packet in the device memory.  */
    rtems_bsdnet_event_receive ( START_TRANSMIT_EVENT,
                                 RTEMS_EVENT_ANY | RTEMS_WAIT,
                                 RTEMS_NO_TIMEOUT,
                                 &events );
    if (debug_level >= DEBUG_STD)
      printf( "ns8390_tx_daemon() - tx event recieved\n" );


#ifdef DEBUG_NE2000
    printk ("ne_tx_daemon\n");
#endif

    /* This daemon handles both uploading data onto the device and
       telling the device to transmit data which has been uploaded.
       These are separate tasks, because while the device is
       transmitting one buffer we will upload another.  */

    /* Don't let the device interrupt us now.  */
    (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_IMR, 0 );

    while (1) {
      struct mbuf *m;

      /* If the device is not transmitting a packet, and we have
         uploaded a packet, tell the device to transmit it.  */
      if (! dp->transmitting && dp->inuse > 0) {
        dp->transmitting = 1;
        ns8390_transmit (dp);
      }

      /* If we don't have any more buffers to send, quit now.  */
      if (ifp->if_snd.ifq_head == NULL) {
        ifp->if_flags &= ~IFF_OACTIVE;
        break;
      }

      /* Allocate a buffer to load data into.  If there are none
         available, quit until a buffer has been transmitted.  */
      if (dp->inuse >= NS8390_TX_BUFS)
        break;

      ++dp->inuse;

      IF_DEQUEUE (&ifp->if_snd, m);
      if (m == NULL)
        panic ("ns8390_tx_daemon");

      ns8390_loadpacket (dp, m);

      /* Check the device status.  It may have finished transmitting
         the last packet.  */
      ns8390_check_status (dp);
    }

    /* Reenable device interrupts.  */
    (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_IMR, NS8390_INTERRUPTS );
  }
}


/* trigger the send of a packet */
/* Start sending an NE2000 packet.  */

static void
ns8390_start (struct ifnet *ifp)
{
  struct ns8390_device  *dp = ifp->if_softc;

  /* Tell the transmit daemon to wake up and send a packet.  */
  rtems_event_send (dp->tx_daemon_tid, START_TRANSMIT_EVENT);
  ifp->if_flags |= IFF_OACTIVE;
}


/* Read data from an NE2000 device.  Read LEN bytes at ADDR, storing
   them into P.  */
static void
ns8390_read_data (struct ns8390_device *dp, int addr, int len, unsigned char *p)
{
  unsigned32            ns8390_port;

  setRegister_f         setReg;
  getData_f             getData;

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg      = dp->setRegister;
  getData     = dp->getData;

  if (debug_level >= DEBUG_STD)
    printf( "ns8390_read_data() - start - addr(%d) len(%d) \n", addr, len );


  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STA );
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RBCR0, len );
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RBCR1, len >> 8 );
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RSAR0, addr );
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RSAR1, addr >> 8 );
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_RD | NS8390_CR_STA );

  while (len > 0) {
    unsigned short d;

    d = (*getData)( ns8390_port );
    *(unsigned16*)p = d;
    p = p + 2;
    len -= 2;
  }

  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_RDC );
}


/* The NE2000 packet receive daemon.  This task is started when the
   NE2000 driver is initialized.  */

static void
ns8390_rx_daemon (void *arg)
{
  struct ns8390_device  *dp = (struct ns8390_device *)arg;
  struct ifnet          *ifp = &dp->arpcom.ac_if;

  unsigned32            ns8390_port;

  setRegister_f         setReg;
  getRegister_f         getReg;
  getData_f             getData;

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg      = dp->setRegister;
  getReg      = dp->getRegister;
  getData     = dp->getData;

  if (debug_level >= DEBUG_STD)
    printf( "ns8390_rx_daemon() - start\n" );

  while (1) {
    rtems_event_set events;

    /* Wait for the interrupt handler to tell us that there is a
       packet ready to receive.  */
    if (debug_level >= DEBUG_STD)
      printf( "ns8390_rx_daemon() - waiting for rx event\n" );

    rtems_bsdnet_event_receive (INTERRUPT_EVENT,
                                RTEMS_WAIT | RTEMS_EVENT_ANY,
                                RTEMS_NO_TIMEOUT,
                                &events);

    if (debug_level >= DEBUG_STD)
      printf( "ns8390_rx_daemon() - rx event recieved\n" );


    /* Don't let the device interrupt us now.  */
    (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_IMR, 0 );

    while (1) {
      unsigned char   startpage, currpage;
      unsigned short  statnext, len;
      unsigned short  a,b;
      int             next;
      struct mbuf     *m;
      unsigned char   *p;
      int             startaddr;
      int             toend;
      struct ether_header *eh;

      startpage = (*getReg)( ns8390_port, NS8390_PAGE0_READ_BNRY );

      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE1 | NS8390_CR_RD_AB );
      currpage = (*getReg)( ns8390_port, NS8390_PAGE1_CURR );
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB );

      ++startpage;
      if (startpage >= NS8390_STOP_PAGE)
        startpage = NS8390_FIRST_RX_PAGE;

      if (startpage == currpage)
        break;

#ifdef DEBUG_NE2000
      printk ("ne_rx_daemon: start page %x; current page %x\n",
	      startpage, currpage);
#endif

    if (debug_level >= DEBUG_STD)
      printf ("ns8390_rx_daemon: start page %x; current page %x\n", startpage, currpage);

      /* Read the buffer header.  This is 1 byte receive status, 1
         byte page of next buffer, 2 bytes length.  */
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STA );
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RBCR0, 4 );
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RBCR1, 0 );
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RSAR0, 0 );
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RSAR1, startpage );
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_RD | NS8390_CR_STA );


      a = (*getData)( ns8390_port );
      b = (*getData)( ns8390_port );

      statnext = (a<<8)|(a>>8);
      len = (b<<8)|(b>>8);


      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_RDC );

#if 0
      next = statnext >> 8;
#else
      next = a >> 8;
#endif

      if (next >= NS8390_STOP_PAGE)
        next = NS8390_FIRST_RX_PAGE;

      /* The first four bytes of the length are the buffer header.  */
      len -= 4;
      startaddr = startpage * NS8390_PAGE_SIZE + 4;

      MGETHDR (m, M_WAIT, MT_DATA);
      MCLGET (m, M_WAIT);
      m->m_pkthdr.rcvif = ifp;

      p = mtod (m, unsigned char *);
      m->m_len = m->m_pkthdr.len = len - sizeof (struct ether_header);

      toend = NS8390_STOP_PAGE * NS8390_PAGE_SIZE - startaddr;
      if (toend < len) 
      {
        ns8390_read_data (dp, startaddr, toend, p);
        p += toend;
        len -= toend;
        startaddr = NS8390_FIRST_RX_PAGE * NS8390_PAGE_SIZE;
      }

      if (len > 0)
        ns8390_read_data (dp, startaddr, len, p);

      eh = mtod (m, struct ether_header *);
      m->m_data += sizeof (struct ether_header);

      if (debug_level >= DEBUG_STD)
      {
        printf ("ns8390_rx_daemon: eh->ether_dhost[] = %X%X%X%X%X%X\n", 
                eh->ether_dhost[0], eh->ether_dhost[1], eh->ether_dhost[2], 
                eh->ether_dhost[3], eh->ether_dhost[4], eh->ether_dhost[5] );
        printf ("ns8390_rx_daemon: eh->ether_shost[] = %X%X%X%X%X%X\n", 
                eh->ether_shost[0], eh->ether_shost[1], eh->ether_shost[2], 
                eh->ether_shost[3], eh->ether_shost[4], eh->ether_shost[5] );
        printf ("ns8390_rx_daemon: eh->ether_type = %4X\n", eh->ether_type );
      }

 

      if (dp->tap != NULL)
        (*dp->tap) (ifp, eh, m);
      else
        ether_input (ifp, eh, m);

      ++dp->eth_stats.rx_packets;

      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_BNRY, next - 1 );
    }

    if (dp->overrun) {
      /* We got a packet overrun.  Reset the device so that we can be
         certain of the current status.  This will lose any packet
         currently being transmitted, but we're probably in trouble
         anyhow.  */
        ns8390_stop (dp);
        ns8390_init (dp);
        dp->overrun = 0;
    }

    /* Reenable device interrupts.  */
    (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_IMR, NS8390_INTERRUPTS );
  }
}




static void
ns8390_stats (struct ns8390_device *dp)
{
  if (debug_level >= DEBUG_STD)
    printf("Some silly fool is trying to call ns8390_stats()\n");

  printf( "TX Packets         : %8d\n", dp->eth_stats.tx_packets );
  printf( "RX Packets         : %8d\n", dp->eth_stats.rx_packets );

  printf( "TX Errors          : %8d\n", dp->eth_stats.tx_errors );
  printf( "RX Errors          : %8d\n", dp->eth_stats.rx_errors );

  printf( "TX Dropped         : %8d\n", dp->eth_stats.tx_dropped );
  printf( "RX Dropped         : %8d\n", dp->eth_stats.rx_dropped );

  printf( "TX Interrupts      : %8d\n", dp->eth_stats.tx_interrupts );
  printf( "RX Interrupts      : %8d\n", dp->eth_stats.rx_interrupts );

  printf( "Any Interrupts     : %8d\n", dp->eth_stats.interrupts );

  printf( "Collisions         : %8d\n", dp->eth_stats.collisions );
  
  printf( "RX Length Errors   : %8d\n", dp->eth_stats.rx_length_errors );
  printf( "RX Overflow Errors : %8d\n", dp->eth_stats.rx_over_errors );
  printf( "RX CRC Errors      : %8d\n", dp->eth_stats.rx_crc_errors );
  printf( "RX Frame Errors    : %8d\n", dp->eth_stats.rx_frame_errors );
  printf( "RX FIFO Errors     : %8d\n", dp->eth_stats.rx_fifo_errors );
  printf( "RX Missed Errors   : %8d\n", dp->eth_stats.rx_missed_errors );

  printf( "TX Aborted Errors  : %8d\n", dp->eth_stats.tx_aborted_errors );
  printf( "TX Carrier Errors  : %8d\n", dp->eth_stats.tx_carrier_errors );
  printf( "TX FIFO Errors     : %8d\n", dp->eth_stats.tx_fifo_errors );
  printf( "TX Heatbeat Errors : %8d\n", dp->eth_stats.tx_heartbeat_errors );
  printf( "TX Window Errors   : %8d\n", dp->eth_stats.tx_window_errors );
  printf( "\n" );
  

  printf( "dp->inuse          : %8d\n", dp->inuse );
  printf( "dp->nextavail      : %8d\n", dp->nextavail );
  printf( "dp->nextsend       : %8d\n", dp->nextsend );
  printf( "dp->transmitting   : %8d\n", dp->transmitting );
  printf( "dp->sendlen[0]     : %8d\n", dp->sendlen[0] );
  printf( "dp->sendlen[1]     : %8d\n", dp->sendlen[1] );
  printf( "dp->overrun        : %8d\n", dp->overrun );
}


/*
 * Driver ioctl handler
 */
static int
ns8390_ioctl (struct ifnet *ifp, int command, caddr_t data)
{
  struct ns8390_device  *dp = ifp->if_softc;
  int           error = 0;


  if (debug_level >= DEBUG_STD)
    printf( "ns8390_ioctl() - start\n" );

  switch (command) {
  case SIOCGIFADDR:
  case SIOCSIFADDR:
    ether_ioctl (ifp, command, data);
    break;

  case SIOCSIFFLAGS:
    switch (ifp->if_flags & (IFF_UP | IFF_RUNNING)) {
    case IFF_RUNNING:
      ns8390_stop (dp);
      break;

    case IFF_UP:
      ns8390_init (dp);
      break;

    case IFF_UP | IFF_RUNNING:
      ns8390_stop (dp);
      ns8390_init (dp);
      break;

    default:
      break;
    }
    break;

  case SIO_RTEMS_SHOW_STATS:
    ns8390_stats (dp);
    break;
    
  /*
   * FIXME: All sorts of multicast commands need to be added here!
   */
  default:
    error = EINVAL;
    break;
  }

  if (debug_level >= DEBUG_STD)
    printf( "ns8390_ioctl() - end\n" );

  return error;
}


/*
 * Find a free driver
 */
struct ns8390_device *
ns8390_get_free_device( void )
{
  struct ifnet      *ifp;
  int           device_number;

  /* Print module version information */
  printf( version );

  for (device_number = 0 ; device_number < NUM_NS8390_DEVICES ; device_number++) 
  {
    ifp = &ns8390_devices[device_number].arpcom.ac_if;
    if (ifp->if_softc == NULL)
      break;
  }
  
  if (device_number >= NUM_NS8390_DEVICES) 
  {
    printf ("Too many NS8390 based drivers.\n");
    return NULL;
  }
  else
  {
    /* Zero out the device structure */
    memset( &ns8390_devices[device_number], 0, sizeof(struct ns8390_device) );

    if (debug_level >= DEBUG_STD)
      printf( "ns8390_get_free_device() - end\n" );
    return &ns8390_devices[device_number];
  }
}



/* External functions called by ns8390.c users */

int
ns8390_driver_attach ( struct rtems_bsdnet_ifconfig *config, struct ns8390_device *dp )
{
  struct ifnet  *ifp = &dp->arpcom.ac_if;
  int       device_number;
  int       mtu;


  if (debug_level >= DEBUG_STD)
    printf( "ns8390_driver_attach() - start\n" );


  /* re-find our device number */
  for (device_number = 0 ; device_number < NUM_NS8390_DEVICES ; device_number++) 
  {
    if (dp == &ns8390_devices[device_number])
      break;
  }


  /*
   * Process options
   */
  if (config->hardware_address) 
  {
    memcpy (dp->arpcom.ac_enaddr, config->hardware_address, ETHER_ADDR_LEN);
  }
  else 
  {
    dp->get_eth_addr( dp, dp->arpcom.ac_enaddr );
  }

  if (config->mtu)
    mtu = config->mtu;
  else
    mtu = ETHERMTU;


  dp->accept_broadcasts = !config->ignore_broadcast;


  /*
   * Set up network interface values
   */
  ifp->if_softc  = dp;
  ifp->if_unit   = device_number + 1;
  ifp->if_name   = "ns";
  ifp->if_mtu    = mtu;
  ifp->if_init   = ns8390_init;
  ifp->if_ioctl  = ns8390_ioctl;
  ifp->if_start  = ns8390_start;
  ifp->if_output = ether_output;
  ifp->if_flags  = IFF_BROADCAST | IFF_SIMPLEX;
  if (ifp->if_snd.ifq_maxlen == 0)
    ifp->if_snd.ifq_maxlen = ifqmaxlen;

  /*
   * Attach the interface
   */
  if_attach (ifp);
  ether_ifattach (ifp);

  printf ("Ethernet address: %X%X%X%X%X%X\n", dp->arpcom.ac_enaddr[0], dp->arpcom.ac_enaddr[1], dp->arpcom.ac_enaddr[2], 
    dp->arpcom.ac_enaddr[3], dp->arpcom.ac_enaddr[4], dp->arpcom.ac_enaddr[5] );


  if (debug_level >= DEBUG_STD)
    printf("ns8390_driver_attach() - end\n");

  return 1;
}




/*
 * Initialize and start the device
 */

/* Initialize and start and NE2000.  */

static void
ns8390_init (void *arg)
{
  struct ns8390_device  *dp = arg;
  struct ifnet          *ifp = &dp->arpcom.ac_if;
  int                   first;
  
  first = 0;
  if (dp->tx_daemon_tid == 0) 
  {
    dp->tx_daemon_tid = rtems_bsdnet_newproc ("NStx", 4096, ns8390_tx_daemon, dp);
    dp->rx_daemon_tid = rtems_bsdnet_newproc ("NSrx", 4096, ns8390_rx_daemon, dp);

    first = 1;
  }

  dp->inuse = 0;
  dp->nextavail = 0;
  dp->nextsend = 0;
  dp->transmitting = 0;

  ns8390_init_hardware (dp);


  ifp->if_flags |= IFF_RUNNING;
  ns8390_startup(dp);
}

/*
 * Initialize the ethernet hardware
 */
void
ns8390_init_hardware( struct ns8390_device *dp )
{
  unsigned8       *hwaddr;
  unsigned32      ns8390_port;

  setRegister_f   setReg;
  getRegister_f   getReg;


  if (debug_level >= DEBUG_STD)
    printf( "ns8390_init_hardware() - start\n" );

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg    = dp->setRegister;
  getReg    = dp->getRegister;
  
  /* Make sure we're on PAGE0, abort any transfer and stop the chip */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_STP | NS8390_CR_RD_AB );
      
  /* set the Data Configuration Register defaults */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_DCR, NS8390_DCR_BOS | NS8390_DCR_WTS | NS8390_DCR_FT_8B | NS8390_DCR_LS );

  /* zero the dma transfer size */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RBCR0, 0 );
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RBCR1, 0 );

  /* set the receive config to monitor */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RCR, NS8390_RCR_MON );

  /* set the transmit config to ENDEC module loopback */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TCR, NS8390_TCR_LOOP_I0 );

#if 1
  /* set the transmit start page */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TPSR, NS8390_FIRST_TX_PAGE );
#endif

#if 0
  /* zero the tx packet buffer sizes */
  dp->tx1 = 0;
  dp->tx2 = 0;
#endif

  /* set start and stop receive addresses */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_PSTART, NS8390_FIRST_RX_PAGE );
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_PSTOP, NS8390_STOP_PAGE );

  /* set internal boundary buffer limit */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_BNRY, NS8390_STOP_PAGE - 1 );

#if 0
  /* set the current packet pointer to the start of the region */
  dp->current_page = dp->rx_start_page;
#endif

  /* disable all interrupts */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, 0xFF );

  /* mask all the chip's interrupts */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_IMR, 0x00 );


  /* Go to PAGE1, abort any transfer and stop the chip */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE1 | NS8390_CR_STP | NS8390_CR_RD_AB );


  /* program supplied hardware address into the chip */
  hwaddr = dp->arpcom.ac_enaddr;
  (*setReg)( ns8390_port, NS8390_PAGE1_PAR0, hwaddr[0] );
  (*setReg)( ns8390_port, NS8390_PAGE1_PAR1, hwaddr[1] );
  (*setReg)( ns8390_port, NS8390_PAGE1_PAR2, hwaddr[2] );
  (*setReg)( ns8390_port, NS8390_PAGE1_PAR3, hwaddr[3] );
  (*setReg)( ns8390_port, NS8390_PAGE1_PAR4, hwaddr[4] );
  (*setReg)( ns8390_port, NS8390_PAGE1_PAR5, hwaddr[5] );


  /* program the multicast address  - to zero because we don't do multicast */
  (*setReg)( ns8390_port, NS8390_PAGE1_MAR0, 0 );
  (*setReg)( ns8390_port, NS8390_PAGE1_MAR1, 0 );
  (*setReg)( ns8390_port, NS8390_PAGE1_MAR2, 0 );
  (*setReg)( ns8390_port, NS8390_PAGE1_MAR3, 0 );
  (*setReg)( ns8390_port, NS8390_PAGE1_MAR4, 0 );
  (*setReg)( ns8390_port, NS8390_PAGE1_MAR5, 0 );
  (*setReg)( ns8390_port, NS8390_PAGE1_MAR6, 0 );
  (*setReg)( ns8390_port, NS8390_PAGE1_MAR7, 0 );

  /* set current page register */
  (*setReg)( ns8390_port, NS8390_PAGE1_CURR, NS8390_FIRST_RX_PAGE );

  /* Return to PAGE0, abort any transfer and stop the chip */
  (*setReg)( ns8390_port, NS8390_PAGE1_CR, NS8390_CR_PAGE0 | NS8390_CR_STP | NS8390_CR_RD_AB );


  if (debug_level >= DEBUG_STD)
    printf( "ns8390_init_hardware() - end\n" );
}


void
ns8390_startup( struct ns8390_device *dp )
{
  unsigned32     ns8390_port;

  setRegister_f  setReg;
  getRegister_f  getReg;


  if (debug_level >= DEBUG_STD)
    printf("ns8390_startup() - start\n");

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg    = dp->setRegister;
  getReg    = dp->getRegister;
  

  /* Final NS8390 configuration and startup */

  /* disable all interrupts */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, 0xFF );

  /* unmask all interrupts */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_IMR, NS8390_INTERRUPTS );

  /* Make sure we're on PAGE0, abort any transfer and startup the chip */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_STA | NS8390_CR_RD_AB );

  /* set the transmit config for normal operation */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TCR, 0 );

  /* set the receive config for normal operation */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RCR, NS8390_RCR_AB );

  if (debug_level >= DEBUG_STD)
    printf("ns8390_startup() - end\n");
}
