/*
 *  Timer Init
 *
 *  This module initializes TIMER 2 for on the MCF5206
 *
 *  Author:
 *    David Fiddes, D.J@fiddes.surfaid.org
 *    http://www.calm.hw.ac.uk/davidf/coldfire/
 *
 *  COPYRIGHT (c) 1989-1998.
 *  On-Line Applications Research Corporation (OAR).
 *  Copyright assigned to U.S. Government, 1994.
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 *
 *  http://www.OARcorp.com/rtems/license.html.
 * 
 *  $Id: timer.c,v 1.8 1998/03/21 15:36:53 joel Exp $
 */

#include <rtems.h>
#include <bsp.h>

#define TRR2_VAL 65530  /* Timer reference register
                         * 1000 ticks @ (25MHz/1)/25 = 1-ms count
                         */

rtems_unsigned32 Timer_interrupts;

rtems_boolean Timer_driver_Find_average_overhead;

extern rtems_isr timerisr( rtems_vector_number vector );


void Timer_initialize( void )
{
  set_vector( timerisr, TIMER2_IRQ_VECTOR, 0);

  sim->icr10 = MCF5206_SIM_ICR_AVEC                   |
               MCF5206_SIM_ICR_IL(TIMER2_IRQ_LEVEL)   |
               MCF5206_SIM_ICR_IP(TIMER2_IRQ_PRIORITY); 

  sim->imr = sim->imr & (~MCF5206_SIM_IMR_T2);

  /* Reset Timer... */
  sim->timer2.tmr = MCF5206_TIMER_TMR_RST;    
  sim->timer2.tmr = MCF5206_TIMER_TMR_CLK_STOP;   
  sim->timer2.tmr = MCF5206_TIMER_TMR_RST;    
    
  sim->timer2.tcn = 0;  /* zero count */
  Timer_interrupts = 0; /* clear timer ISR count */
  sim->timer2.ter = MCF5206_TIMER_TER_REF | MCF5206_TIMER_TER_CAP; /* clear pending */
    
    
          
  /* Setup Timer... */
  sim->timer2.trr = TRR2_VAL - 1; /* subtract 1 for timer overcount */
  sim->timer2.tmr = MCF5206_TIMER_TMR_PS( MCF5206_CLOCK_SPEED/1000000 )    |
                    MCF5206_TIMER_TMR_CE_NONE   |
                    MCF5206_TIMER_TMR_ORI       |
                    MCF5206_TIMER_TMR_FRR       |
                    MCF5206_TIMER_TMR_RST;
          

  sim->timer2.tmr = sim->timer2.tmr | MCF5206_TIMER_TMR_CLK_MSCLK;

}

/*
 *  The following controls the behavior of Read_timer().
 *
 *  FIND_AVG_OVERHEAD *  instructs the routine to return the "raw" count.
 *
 *  AVG_OVEREHAD is the overhead for starting and stopping the timer.  It
 *  is usually deducted from the number returned.
 *
 *  LEAST_VALID is the lowest number this routine should trust.  Numbers
 *  below this are "noise" and zero is returned.
 */

#define AVG_OVERHEAD      0  /* It typically takes 2.0 microseconds */
                             /* (Y countdowns) to start/stop the timer. */
                             /* This value is in microseconds. */
#define LEAST_VALID       1  /* Don't trust a clicks value lower than this */

/*
 * Return timer value in microsecond units
 */
int Read_timer( void )
{
  rtems_unsigned16 clicks;
  rtems_unsigned32 total;

  /*
   *  Read the timer and see how many clicks it has been since counter
   *  rolled over.
   */

  clicks = sim->timer2.tcn;
  
  /* Stop Timer... */
  sim->timer2.tmr = MCF5206_TIMER_TMR_CLK_STOP |
                    MCF5206_TIMER_TMR_RST;    

  /*
   *  Total is calculated by taking into account the number of timer overflow
   *  interrupts since the timer was initialized and clicks since the last
   *  interrupts.
   */

  total = (Timer_interrupts * TRR2_VAL) + clicks;

  if ( Timer_driver_Find_average_overhead == 1 )
    return total;          /* in XXX microsecond units */

  if ( total < LEAST_VALID )
    return 0;            /* below timer resolution */

  /*
   *  Return the count in microseconds
   */

  return (total - AVG_OVERHEAD);
}


/*
 *  Empty function call used in loops to measure basic cost of looping
 *  in Timing Test Suite.
 */

rtems_status_code Empty_function(void)
{
  return RTEMS_SUCCESSFUL;
}

void Set_find_average_overhead(
  rtems_boolean find_flag
)
{
  Timer_driver_Find_average_overhead = find_flag;
}
